VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CCrypto"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'****************************************************
'* CCrypto.cls                                      *
'* advanced Xor encryption class                    *
'* provides a nearly 100% secure encryption that is *
'* uncrackable by modern computers due to the use   *
'* of a OneTimePad and a high precision pseudo      *
'* random creator                                   *
'* for further details please see the documentation *
'* Programmed: Konrad Rudolph (Mad Rat Pro.)        *
'* Last Change: 23.04.2003 22:14                    *
'* Version: 1.0.0                                   *
'****************************************************
Option Explicit

Private Type SHA_Hash
    Hash(7) As Long         ' SHA-1 hash
End Type

Private Type CryptoHeader
    CheckKey    As SHA_Hash ' SHA-1 checksum
    TimeStamp   As Long     ' date of encryption
End Type
'----------------------------------------------------

Private mInfo       As CryptoHeader     ' crypto information for current session
Private mlInfoLen   As Long             ' length of CryptHeader structure
Private msKey       As String           ' current key

Private mShaInst    As CSHA             ' SHA class instance
Private mRndInst    As CRndAccu         ' random generators wrapper instance

Private mbSessOpen  As Boolean          ' session opened?
'----------------------------------------------------
Private Sub Class_Initialize()
    mlInfoLen = LenB(mInfo)             ' assign length of the CryptoHeader struct
End Sub
'----------------------------------------------------

'----------------------------------------------------
' Property Let Key As String
' assigns a new key to the current session
' NO Property Get for security purpose (ok, ridiculous...)
'----------------------------------------------------
Public Property Let Key(Value As String)
    If Not mbSessOpen Then _
        Exit Property
    
    msKey = Value                       ' assign key
    Call HashKey                        ' hash key
End Property

'----------------------------------------------------
' CreateSession
' generates a new crypto session
'----------------------------------------------------
Public Sub CreateSession()
    Call CloseSession
    
    Set mShaInst = New CSHA
    Set mRndInst = New CRndAccu
    mRndInst.Count = 8                  ' number of SHA registers in one hash
    
    mInfo.TimeStamp = GetTimeStamp      ' get a time stamp for the session
    mbSessOpen = True                   ' mark the session opened
End Sub

'----------------------------------------------------
' CloseSession
' closes the currently open crypto session
'----------------------------------------------------
Public Sub CloseSession()
    Dim I   As Long
    
    If Not mbSessOpen Then _
        Exit Sub
    
    For I = 1 To Len(msKey)
        Mid$(msKey, I, 1) = "@"
    Next I
    msKey = ""
    
    mbSessOpen = False
    Set mShaInst = Nothing
    Set mRndInst = Nothing
End Sub

'----------------------------------------------------
' EncodeBytes As Byte()
' encodes a one-dimension bytearray
' - IN InputArr() As Byte: array to encode
' - WipeClean As Boolean: if set, the OTP is overwritten
' with null-bytes after writing. This ensures more security
' but is quite slow and should normally not be needed
' returns the encoded byte array
'----------------------------------------------------
Public Function EncodeBytes( _
  InputArr() As Byte, _
  Optional WipeClean As Boolean = True _
) As Byte()
    Dim OTP()       As Byte
    Dim Cypher()    As Byte
    Dim LB          As Long, _
        UB          As Long, _
        DiffB       As Long
    Dim I           As Long
    
    Dim Byte1       As Byte, _
        Byte2       As Byte, _
        C_Byte      As Byte
    Dim lpIn        As Long, _
        lpKey       As Long, _
        lpOut       As Long
    
    If Not mbSessOpen Then _
        Exit Function
    
    LB = LBound(InputArr)
    UB = UBound(InputArr)
    DiffB = UB - LB
    
    OTP = CreateOTP(DiffB)
    ReDim Cypher(LB To UB + mlInfoLen) As Byte
    
    lpIn = VarPtr(InputArr(LB))
    lpKey = VarPtr(OTP(0))
    lpOut = VarPtr(Cypher(LB))
    
    Call Poke(lpOut, mInfo, mlInfoLen)
    
    lpOut = lpOut + mlInfoLen
    
    For I = 0 To DiffB
        Call Peek(Byte1, lpIn + I, 1)
        Call Peek(Byte2, lpKey + I, 1)
        
        C_Byte = Byte1 Xor Byte2
        
        Call Poke(lpOut + I, C_Byte, 1)
    Next I
    
    EncodeBytes = Cypher
    
    If WipeClean Then
        For I = 0 To UBound(OTP)
            OTP(I) = 0
        Next I
    End If
    
    Erase OTP
End Function

'----------------------------------------------------
' DecodeBytes As Byte()
' decodes a one-dimension bytearray
' - IN InputArr() As Byte: array to decode
' - WipeClean As Boolean: if set, the OTP is overwritten
' with null-bytes after writing. This ensures more security
' but is quite slow and should normally not be needed
' returns the decoded byte array
'----------------------------------------------------
Public Function DecodeBytes( _
  InputArr() As Byte, _
  Optional WipeClean As Boolean = True _
) As Byte()
    Dim OTP()       As Byte
    Dim Plain()     As Byte
    Dim LB          As Long, _
        UB          As Long, _
        DiffB       As Long
    Dim I           As Long
    
    Dim Byte1       As Byte, _
        Byte2       As Byte, _
        C_Byte      As Byte
    Dim lpIn        As Long, _
        lpKey       As Long, _
        lpOut       As Long
    
    If Not mbSessOpen Then _
        Exit Function
    
    LB = LBound(InputArr)
    UB = UBound(InputArr)
    DiffB = UB - LB - mlInfoLen
    
    lpIn = VarPtr(InputArr(LB))
    Call Peek(mInfo, lpIn, mlInfoLen)
    
    If Not HashKey(Check:=True) Then
        ReDim Plain(0) As Byte
        DecodeBytes = Plain
        Exit Function
    End If
    
    OTP = CreateOTP(DiffB)
    ReDim Plain(LB To UB - mlInfoLen) As Byte
    
    lpIn = lpIn + mlInfoLen
    lpKey = VarPtr(OTP(0))
    lpOut = VarPtr(Plain(LB))
    
    For I = 0 To DiffB
        Call Peek(Byte1, lpIn + I, 1)
        Call Peek(Byte2, lpKey + I, 1)
        
        C_Byte = Byte1 Xor Byte2
        
        Call Poke(lpOut + I, C_Byte, 1)
    Next I
    
    DecodeBytes = Plain
    
    If WipeClean Then
        For I = 0 To UBound(OTP)
            OTP(I) = 0
        Next I
    End If
    
    Erase OTP
End Function

'----------------------------------------------------
' EncodeString As String
' String-wrapper for EncodeBytes
'----------------------------------------------------
Public Function EncodeString( _
  InputStr As String, _
  Optional WipeClean As Boolean = True _
) As String
    EncodeString = StrConv(EncodeBytes(StrConv(InputStr, vbFromUnicode), _
      WipeClean), vbUnicode)
End Function

'----------------------------------------------------
' DecodeString As String
' String-wrapper for DecodeBytes
'----------------------------------------------------
Public Function DecodeString( _
  InputStr As String, _
  Optional WipeClean As Boolean = True _
) As String
    DecodeString = StrConv(DecodeBytes(StrConv(InputStr, vbFromUnicode), _
      WipeClean), vbUnicode)
End Function
'----------------------------------------------------

'----------------------------------------------------
' Permutate As Long()
' mixes and hashes the key and a timestamp
' to create a SHA-hash used as random seed
' - IN Key() As Byte: the key in a byte array
' - IN TimeStamp() AS Byte: the timestamp in a byte array
' returns a Long array containing the hash
'----------------------------------------------------
Private Function Permutate( _
  Key() As Byte, _
  TimeStamp() As Byte _
) As Long()
    Dim tmpBuff()   As Byte
    Dim I           As Long
    
    ReDim tmpBuff(0 To UBound(Key) + UBound(TimeStamp) + 1) As Byte
    
    tmpBuff(0) = TimeStamp(0)
    tmpBuff(1) = TimeStamp(1)
    For I = 0 To UBound(Key)
        tmpBuff(I + 2) = Key(I)
    Next I
    tmpBuff(UBound(tmpBuff) - 1) = TimeStamp(2)
    tmpBuff(UBound(tmpBuff)) = TimeStamp(3)
    
    Permutate = mShaInst.SHA256(StrConv(tmpBuff, vbUnicode))
End Function

'----------------------------------------------------
' ToHash As SHA_Hash
' converts a Long array into the SHA_Hash struct
' (non-API solution)
'----------------------------------------------------
Private Function ToHash(Hash() As Long) As SHA_Hash
    With ToHash
        .Hash(0) = Hash(0)
        .Hash(1) = Hash(1)
        .Hash(2) = Hash(2)
        .Hash(3) = Hash(3)
        .Hash(4) = Hash(4)
        .Hash(5) = Hash(5)
        .Hash(6) = Hash(6)
        .Hash(7) = Hash(7)
    End With
End Function

'----------------------------------------------------
' HashEqual As Boolean
' compares two hashes and returns if they are equal
' - IN Hash1 As SHA_Hash: first hash value
' - IN Hash2 As SHA_Hash: second hash value
'----------------------------------------------------
Private Function HashEqual( _
  Hash1 As SHA_Hash, _
  Hash2 As SHA_Hash _
) As Boolean
    With Hash1
        If .Hash(0) <> Hash2.Hash(0) Then _
            Exit Function
        If .Hash(1) <> Hash2.Hash(1) Then _
            Exit Function
        If .Hash(2) <> Hash2.Hash(2) Then _
            Exit Function
        If .Hash(3) <> Hash2.Hash(3) Then _
            Exit Function
        If .Hash(4) <> Hash2.Hash(4) Then _
            Exit Function
        If .Hash(5) <> Hash2.Hash(5) Then _
            Exit Function
        If .Hash(6) <> Hash2.Hash(6) Then _
            Exit Function
        If .Hash(7) <> Hash2.Hash(7) Then _
            Exit Function
    End With
    
    HashEqual = True
End Function

'----------------------------------------------------
' CreateOTP As Byte()
' fills a byte array of a given size with random
' bytes
' - IN Size As Long: size of the OTP
' returns the OTP array
'----------------------------------------------------
Private Function CreateOTP(Size As Long) As Byte()
    Dim tmpOTP()    As Byte
    Dim I           As Long
    
    ReDim tmpOTP(0 To Size) As Byte
    
    For I = 0 To Size
        tmpOTP(I) = CByte(mRndInst.Random(0, 255))
    Next I
    
    CreateOTP = tmpOTP
    
    Erase tmpOTP
End Function

'----------------------------------------------------
' HashKey As Boolean
' fulfills two purposes:
' . it hashes the current key and timestamp an creates a
'   checksum for the key to test if the given decryption key
'   is right
'   it then initialises the random generators with the
'   permutated key and timestamp
' . on demand it checks whether the current key corresponds
'   to the stored checksum
' - IN Check As Boolean: identifies whether a key check shall
'   be performed or not
' if check is to be performed the function returns accordance
' if not it returns False
'----------------------------------------------------
Private Function HashKey( _
  Optional ByVal Check As Boolean = False _
) As Boolean
    Dim tmpKey()    As Byte
    Dim lKey()      As Long
    Dim I           As Long
    Dim Hash        As SHA_Hash
    
    tmpKey = StrConv(msKey, vbFromUnicode)
    Hash = ToHash(mShaInst.SHA256(StrConv(tmpKey, vbUnicode)))
    
    If Check Then
        If HashEqual(Hash, mInfo.CheckKey) Then _
            HashKey = True _
        Else _
            Exit Function
    Else
        mInfo.CheckKey = Hash
    End If
    
    lKey = Permutate(tmpKey, Ptr2Str(VarPtr(mInfo.TimeStamp), 4))
    
    For I = 0 To UBound(lKey)
        mRndInst.Seed = lKey(I)
    Next I
End Function
